;;; Directory Local Variables
;;; For more information see (info "(emacs) Directory Variables")
;;; Match project coding conventions

((c-mode . ((indent-tabs-mode . t)
            (show-trailing-whitespace . t)
            (c-basic-offset . 8)))
 (json-mode . ((js-indent-level 4))))
