// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * STATICd - vty header
 * Copyright (C) 2018 Cumulus Networks, Inc.
 *               Donald Sharp
 */
#ifndef __STATIC_VTY_H__
#define __STATIC_VTY_H__

#ifdef __cplusplus
extern "C" {
#endif

void static_vty_init(void);

#ifdef __cplusplus
}
#endif

#endif
