// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * qpb.c
 *
 * @copyright Copyright (C) 2016 Sproute Networks, Inc.
 *
 * @author Avneesh Sachdev <avneesh@sproute.com>
 */

/*
 * Main file for the qpb library.
 */

#include "config.h"
#include "xref.h"

XREF_SETUP();
