// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * EIGRP Finite State Machine (DUAL).
 * Copyright (C) 2013-2014
 * Authors:
 *   Donnie Savage
 *   Jan Janovic
 *   Matej Perina
 *   Peter Orsag
 *   Peter Paluch
 */

#ifndef _ZEBRA_EIGRP_FSM_H
#define _ZEBRA_EIGRP_FSM_H

extern int eigrp_fsm_event(struct eigrp_fsm_action_message *msg);


#endif /* _ZEBRA_EIGRP_DUAL_H */
