// SPDX-License-Identifier: GPL-2.0-or-later
/* VLAN (802.1q) common header.
 * Copyright (C) 2016, 2017 Cumulus Networks, Inc.
 */

#ifndef __VLAN_H__
#define __VLAN_H__

#ifdef __cplusplus
extern "C" {
#endif

/* VLAN Identifier */
typedef uint16_t vlanid_t;
#define VLANID_MAX 4095

#ifdef __cplusplus
}
#endif

#endif /* __VLAN_H__ */
