// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * EIGRP VTY Interface.
 * Copyright (C) 2013-2016
 * Authors:
 *   Donnie Savage
 *   Jan Janovic
 *   Matej Perina
 *   Peter Orsag
 *   Peter Paluch
 *   Frantisek Gazo
 *   Tomas Hvorkovy
 *   Martin Kontsek
 *   Lukas Koribsky
 */

#ifndef _QUAGGA_EIGRP_VTY_H
#define _QUAGGA_EIGRP_VTY_H

/* Prototypes. */
extern void eigrp_vty_init(void);
extern void eigrp_vty_show_init(void);

#endif /* _Quagga_EIGRP_VTY_H_ */
