// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * pbr memory code.
 * Copyright (C) 2018 Cumulus Networks, Inc.
 *               Donald Sharp
 */
#ifndef __PBR_MEMORY_H__

DECLARE_MGROUP(PBRD);

#endif
