// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * PBR memory code.
 * Copyright (C) 2018 Cumulus Networks, Inc.
 *               Donald Sharp
 */
#include <zebra.h>

#include <memory.h>

#include "pbrd/pbr_memory.h"


DEFINE_MGROUP(PBRD, "pbrd");
