// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * PIM for Quagga
 * Copyright (C) 2008  Everton da Silva Marques
 */

#ifndef PIM_SIGNALS_H
#define PIM_SIGNALS_H

#include "sigevent.h"
extern struct frr_signal_t pimd_signals[];

#endif /* PIM_SIGNALS_H */
