// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * Copyright (C) 2020 Volta Networks, Inc
 *                     Brady Johnson
 */

#ifndef _PATH_PCEP_CLI_H_
#define _PATH_PCEP_CLI_H_


/* PCEP CLI Functions */
void pcep_cli_init(void);

#endif // _PATH_PCEP_CLI_H_
