// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * EIGRP YANG Functions.
 * Copyright (C) 2019
 * Authors:
 *   Donnie Savage
 */

#ifndef _EIGRP_YANG_H_
#define _EIGRP_YANG_H_

/*Prototypes*/

/* eigrp_northbound.c */
extern const struct frr_yang_module_info frr_eigrpd_info;

#endif /*EIGRP_YANG_H_ */
