// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * IS-IS Rout(e)ing protocol - isis_routemap.h
 *
 * Copyright (C) 2013-2015 Christian Franke <chris@opensourcerouting.org>
 */
#ifndef ISIS_ROUTEMAP_H
#define ISIS_ROUTEMAP_H

void isis_route_map_init(void);

#endif
