// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * EIGRP SNMP Support.
 * Copyright (C) 2013-2014
 * Authors:
 *   Donnie Savage
 *   Jan Janovic
 *   Matej Perina
 *   Peter Orsag
 *   Peter Paluch
 */


#ifndef _ZEBRA_EIGRP_SNMP_H
#define _ZEBRA_EIGRP_SNMP_H

extern void eigrp_snmp_init(void);


#endif /* _ZEBRA_EIGRP_SNMP_H */
