// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * Test wrappers common header file
 *
 * Copyright (C) 2015 by David Lamparter,
 *                    for Open Source Routing./ NetDEF, Inc.
 *
 * This file is part of Quagga
 */

#ifndef _QUAGGA_TESTS_H
#define _QUAGGA_TESTS_H

extern void test_init(void);
extern void test_init_cmd(void);

#endif /* _QUAGGA_TESTS_H */
