// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * VTY library for PBR
 * Copyright (C) 2018 Cumulus Networks, Inc.
 *               Donald Sharp
 */
#ifndef __PBR_VTY_H__
#define __PBR_VTY_H__

extern void pbr_vty_init(void);
#endif
